//==============================================================================
//	
//	Copyright (c) 2015-
//	Authors:
//	* Joachim Klein <klein@tcs.inf.tu-dresden.de> (TU Dresden)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package explicit;

import java.util.BitSet;

import parser.ast.Expression;
import prism.PrismException;


/**
 * Combines two (compatible) ModelExpressionTransformation.
 * <br>
 * The inner transformation is applied first, with the result
 * serving as the input for the second, outer transformation.
 *
 * @param <OriginalModel> The input type for the inner transformation
 * @param <IntermediateModel> The output type for the inner and the input type for the outer transformation
 * @param <TransformedModel> The output type for the outer transformation
 */
public class ModelExpressionTransformationNested<OriginalModel extends Model, IntermediateModel extends Model, TransformedModel extends Model> implements
       ModelExpressionTransformation<OriginalModel, TransformedModel> {

	/** The inner transformation */
	protected ModelExpressionTransformation<OriginalModel, IntermediateModel> innerTransformation;
	/** The outer transformation */
	protected ModelExpressionTransformation<IntermediateModel,TransformedModel> outerTransformation;

	/**
	 * Constructor.
	 * <br>
	 * The output (transformed) model of the inner transformation
	 * has to be the input (original) model of the outer transformation.
	 *
	 * @param innerTransformation the inner transformation
	 * @param outerTransformation the outer transformation
	 */
	public ModelExpressionTransformationNested(ModelExpressionTransformation<OriginalModel, IntermediateModel> innerTransformation,
			ModelExpressionTransformation<IntermediateModel, TransformedModel> outerTransformation) throws PrismException {

		this.innerTransformation = innerTransformation;
		this.outerTransformation = outerTransformation;

		if (innerTransformation.getTransformedModel() != outerTransformation.getOriginalModel()) {
			throw new PrismException("Trying to nest unrelated ModelExpressionTransformations.");
		}
	}

	@Override
	public TransformedModel getTransformedModel() throws PrismException {
		// the transformed model for the nested transformation
		// is the transformed model for the outer transformation
		return outerTransformation.getTransformedModel();
	}

	@Override
	public BitSet getTransformedStatesOfInterest() throws PrismException {
		// the transformed states of interest for the nested transformation
		// are the transformed states of interest for the outer transformation
		return outerTransformation.getTransformedStatesOfInterest();
	}

	@Override
	public StateValues projectToOriginalModel(StateValues svTransformedModel) throws PrismException {
		// first, get the StateValues in the intermediate model
		StateValues svIntermediate = outerTransformation.projectToOriginalModel(svTransformedModel);
		// then, in the original model
		StateValues svOriginal = innerTransformation.projectToOriginalModel(svIntermediate);
		return svOriginal;
	}

	@Override
	public Expression getTransformedExpression() throws PrismException {
		// the transformed expression for the nested transformation
		// is the transformed expression for the outer transformation
		return outerTransformation.getTransformedExpression();
	}

	@Override
	public OriginalModel getOriginalModel() {
		// the original model for the nested transformation
		// is the original model for the inner transformation
		return innerTransformation.getOriginalModel();
	}

	@Override
	public Expression getOriginalExpression() throws PrismException {
		// the original expression for the nested transformation
		// is the original expression for the inner transformation
		return innerTransformation.getOriginalExpression();
	}
}
